### A Pluto.jl notebook ###
# v0.17.5

using Markdown
using InteractiveUtils

# This Pluto notebook uses @bind for interactivity. When running this notebook outside of Pluto, the following 'mock version' of @bind gives bound variables a default value (instead of an error).
macro bind(def, element)
    quote
        local iv = try Base.loaded_modules[Base.PkgId(Base.UUID("6e696c72-6542-2067-7265-42206c756150"), "AbstractPlutoDingetjes")].Bonds.initial_value catch; b -> missing; end
        local el = $(esc(element))
        global $(esc(def)) = Core.applicable(Base.get, el) ? Base.get(el) : iv(el)
        el
    end
end

# ╔═╡ a7b272bc-e8f5-11eb-38c6-8f61fea9941c
using CSV, DataFrames, PlotlyJS, PlutoUI, Random, Statistics

# ╔═╡ 131466aa-d30d-4f75-a99f-13f47e1c7956
using LinearAlgebra: dot, norm, norm1, norm2

# ╔═╡ 7f8a82ea-1690-490c-a8bc-3c1f9556af2e
using Distributions: Uniform

# ╔═╡ 6fef79fe-aa3c-497b-92d3-6ddd87b6c26d
begin
	_check_complete(complete) = complete ? "✅" : "❌"
	
	md"""
	# Setup
	
	This section loads and installs all the packages. You should be setup already from assignment 1, but if not please read and follow the `instructions.md` for further details.
	"""
end

# ╔═╡ ac9b31de-a4f0-4cb6-82a7-890436ff1bed
import Colors: @colorant_str

# ╔═╡ 5924a9a5-88c7-4751-af77-4a217dfdc15f
md"""
### !!!IMPORTANT!!!

Insert your details below. You should see a green checkmark.
"""

# ╔═╡ 9d38b22c-5893-4587-9c33-82d846fd0728
student = (name="NAME as in eclass", email="UofA Email", ccid="CCID", idnumber=0)

# ╔═╡ 72d6b53f-f20d-4c05-abd4-d6fb4c997795
let
	def_student = (name="NAME as in eclass", email="UofA Email", ccid="CCID", idnumber=0)
	if length(keys(def_student) ∩ keys(student)) != length(keys(def_student))
		md"You don't have all the right entries! Make sure you have `name`, `email`, `ccid`, `idnumber`. ❌"
	elseif any(getfield(def_student, k) == getfield(student, k) for k in keys(def_student))
		md"You haven't filled in all your details! ❌"
	elseif !all(typeof(getfield(def_student, k)) === typeof(getfield(student, k)) for k in keys(def_student))
		md"Your types seem to be off: `name::String`, `email::String`, `ccid::String`, `idnumber::Int`"
	else
		md"Welcome $(student.name)! ✅"
	end
end

# ╔═╡ e10ffb7d-ede9-4cec-8908-836cd1ee3ae1
md"""
Important Note: You should only write code in the cells that has: """


# ╔═╡ a095e895-cabb-4e3a-a027-ddd6aa356e41
#### BEGIN SOLUTION


#### END SOLUTION

# ╔═╡ c8ceaf7a-d740-4123-8b26-e652dd17a86e
md"""
# Distance Metrics

Here we define some functions for commonly used distance metrics, to later allows us to measure errors.

"""

# ╔═╡ b3edc600-6670-4ae6-af7a-337e0600390e
begin
	RMSE(x̂, x) = sqrt(mean(abs2.(x̂ .- x))) # abs2 is equivalent to squaring, but faster and better numerically. 
	l2_error(x̂, x) = norm2(x̂ .- x)
	l1_error(x̂, x) = norm1(x̂ .- x)
end

# ╔═╡ ac180a6e-009f-4db2-bdfc-a9011dc5eb7b
md"""
# Abstract type Regressor

This is the basic Regressor interface. For the methods below we will be specializing the `predict(reg::Regressor, x::Number)`, and `epoch!(reg::Regressor, args...)` functions. Notice the `!` character at the end of epoch. This is a common naming convention in Julia to indicate that this is a function that modifies its arguments.


"""

# ╔═╡ 999ce1f2-fd11-4584-9c2e-bb79585d11f7
"""
	Regressor

Abstract Type for regression algorithms. Interface includes `predict` and an `epoch!`. In this notebook, we will only be using single variate regression.
- `predict(reg::Regressor, X::Number)`: return a prediction of the target given the feature `x`.
- `epoch!(reg::Regressor, X::AbstractVector, Y::AbstractVector)`: trains using the features `X` and regression targets `Y`.
"""
abstract type Regressor end # assume linear regression

# ╔═╡ 80f4168d-c5a4-41ef-a542-4e6d7772aa80
predict(reg::Regressor, x::Number) = Nothing

# ╔═╡ 65505c23-5096-4621-b3d4-d8d5ed339ad5
predict(reg::Regressor, X::AbstractVector) = [predict(reg, x) for x in X]

# ╔═╡ 67a8db8a-365f-4912-a058-d61648ac096e
epoch!(reg::Regressor, X::AbstractVector, Y::AbstractVector) = nothing

# ╔═╡ 7e083787-7a16-4064-97d9-8e2fca6ed686
md"""
# Baselines

In this section we will define the:
- `MeanRegressor`: Predict the mean of the training set.
- `RandomRegressor`: Predict `w*x` where `w` is sampled from a random normal distribution.

All the following baselines assume one dimension
"""

# ╔═╡ 5d5fe2d8-98ea-4ee1-b2e2-354eefaf8f77
md"""
## MeanRegressor
"""

# ╔═╡ eb152f9f-908f-4e5b-9642-f7314eb66e09
begin
	"""
		MeanRegressor()
		
	Predicts the mean value of the regression targets passed in through `epoch!`.
	"""
	mutable struct MeanRegressor <: Regressor
		μ::Float64
	end
	MeanRegressor() = MeanRegressor(0.0)
	predict(reg::MeanRegressor, x::Number) = reg.μ
	epoch!(reg::MeanRegressor, X::AbstractVector, Y::AbstractVector) = reg.μ = mean(Y)
end

# ╔═╡ a7b7bdaf-f4fd-40fb-a704-a584a93920f2
md"""
## RandomRegressor
"""

# ╔═╡ 0a90ceec-fc50-41fb-bf8b-e96ed677a5c3
begin
	"""
		RandomRegressor
	
	Predicts `w*x` where `w` is sambled from a normal distribution.
	"""
	struct RandomRegressor <: Regressor # random weights
		w::Float64
	end
	RandomRegressor() = RandomRegressor(randn())
	predict(reg::RandomRegressor, x::Number) = reg.w*x
end

# ╔═╡ 76c5de40-a0de-4d90-82f0-4f6c941499d9
html"""<h1 id="q3abc"> Q3 a,b,c: Gradient Descent Regressors </h1>"""

# ╔═╡ e8f3981d-545c-4e35-9066-69fa4f78dbce
md"""
In this section you will be implementing two gradient descent regressors, assuming $$p(y | x)$$ is Gaussian with the update given in the assignment pdf.  First we will create a Gaussian regressor, and then use this to build our two new GD regressors. You can test your algorithms in the [experiment section](#experiment).

All the Gaussian Regressors will have data:
- `w::Float64` which is the parameter we are learning.
"""

# ╔═╡ 6ef4d625-fdf6-4f11-81f1-b242b2195e8b
abstract type GaussianRegressor <: Regressor end

# ╔═╡ 579da3a3-0e4e-4ba2-8f44-2657229994e3
predict(reg::GaussianRegressor, x::Float64) = reg.w * x

# ╔═╡ 67274c5b-2ee3-4865-a5e1-799db543d0c7
predict(reg::GaussianRegressor, X::Vector{Float64}) = reg.w .* X

# ╔═╡ 57c4dbf9-b7ba-44e9-8120-0e1a85896976
function probability(reg::GaussianRegressor, x, y)
end

# ╔═╡ add25a22-7423-4077-b0b0-f71eed2d2e20
begin
	md"""
	The stochastic regressor will be implemented via the stochastic gradient rule
	
	```math
	\begin{align*}
	w_{t+1} = w_t - \eta (x_i w_t - y_i) x_i
	\end{align*}
	```
	
	by starting from an initial $$w_0 = 0$$ and doing multiple epochs over the dataset. Each epoch corresponds to iterating once over the entire dataset, in a random order. Here you only have to implement the function that does one epoch. It will be called multiple times in `train!`, which is a function provided for you described in the section on Training the Models. After you implement your algorithm, you should test it in the experiments section to answer Q3 b. Note: to obtain the length of an array `arr`, you can use `length(arr)`. If you see an error that says "MethodError: no method matching...", then this means you have an error in your implementation.
	"""
end

# ╔═╡ b0160dc0-35af-4750-9ac6-d9018ce89ea9
begin
	mutable struct StochasticRegressor <: GaussianRegressor
		w::Float64
		η::Float64
	end
	StochasticRegressor(η::Float64) = StochasticRegressor(0.0, η)
end

# ╔═╡ adedf31d-945f-491a-912a-bef9b03f6665
# Hint: Checkout the function randperm
function epoch!(reg::StochasticRegressor, 
		         X::AbstractVector{Float64}, 
		         Y::AbstractVector{Float64})
	# Your code here is for question 3(a)
	# For question 3(b), you will move to the experiment section. Remember to change the stepsize (η::Float64) 
	#### BEGIN SOLUTION


	#### END SOLUTION
end

# ╔═╡ 9b69b182-ee8d-485e-ad6b-1f868874652c
begin
	md"""
	The Minibatch regressor will be implemented via the gradient rule for a minibatch `j` with indicies for a batch defined by the set  $$\mathcal{I}$$

	```math
	\begin{align*}
	g_t^j &= \sum_{i\in \mathcal{I}_j} (x_i w_t - y_i) x_i \\
	w_{t+1} &= w_t - \eta g^j_t.
	\end{align*}
	```

	If no batch size `b` is specified, then your implementation should run Batch Gradient Descent. Once again you need to implement the `epoch!` function, which will be called elsewhere by `train!`. 
	"""
end

# ╔═╡ 4d422513-34e3-4d04-8ff1-c5165d953342
begin
	mutable struct BatchRegressor <: GaussianRegressor
		w::Float64
		η::Float64
		b::Union{Int, Nothing}
	end
	BatchRegressor(η, b=nothing) = BatchRegressor(0.0, η, b)
end

# ╔═╡ e5d78429-55f4-4a1a-831f-dbcb53c6a0f6
function epoch!(reg::BatchRegressor,
		        X::AbstractVector{Float64},
		        Y::AbstractVector{Float64})
	# Your code here is for question 3(c)
	# You will need to move to the experiment section to test your implementation
	#### BEGIN SOLUTION


	#### END SOLUTION
end

# ╔═╡ 2aa8bd30-4b65-4e89-9dd5-5333efbbda3f
md"""
# Stepsize Heuristic
"""

# ╔═╡ 51bc41b4-b27f-4e60-8dba-70783c60c1c2
begin
	mutable struct StochasticHeuristicRegressor <: GaussianRegressor
		w::Float64
	end
	StochasticHeuristicRegressor() = 	
		StochasticHeuristicRegressor(0.0)
end

# ╔═╡ 00fcafa5-308c-4895-ad52-961772348125
# Hint: Checkout the function randperm
function epoch!(reg::StochasticHeuristicRegressor, 
		         X::AbstractVector{Float64}, 
		         Y::AbstractVector{Float64})
	# Your code here is for 3(d)
	#### BEGIN SOLUTION


	#### END SOLUTION
end

# ╔═╡ 7d6027af-6bfe-4cca-8b81-3e8fb12a6f79
begin
	mutable struct BatchHeuristicRegressor <: GaussianRegressor
		w::Float64
		b::Union{Int, Nothing}
	end
	BatchHeuristicRegressor(b=nothing) = BatchHeuristicRegressor(0.0, b)
end

# ╔═╡ 297b3bd1-29cc-4a96-bf2c-6305f8d375e4
function epoch!(reg::BatchHeuristicRegressor,
		        X::AbstractVector{Float64},
		        Y::AbstractVector{Float64})
	# Your code here is for question 3(e)
	#### BEGIN SOLUTION


	#### END SOLUTION
end

# ╔═╡ 06be63f6-577c-4117-8dac-bba8bba8b9ce
begin
	__check_stoch_reg = let
		sgr = StochasticRegressor(0.0, 0.1)
		epoch!(sgr, [1.0, 1.0], [1.0, 1.0])
		predict(sgr, 1.0) ≈ 0.19
	end

	HTML("<h3 id=q3ab> Stochastic Regressor $(_check_complete(__check_stoch_reg)) </h3>")
end

# ╔═╡ debd1501-2fae-431c-85cf-bea549a4bba4
begin
	__check_batch_reg = let
		X, Y = [1.0, 2.0, 3.0], [1.0, 0.2, 0.1]
		bgr = BatchRegressor(0.1*3)
		epoch!(bgr, X, Y)
		predict(bgr, 1.0) ≈ 0.17
	end

	__check_mb = let
		X, Y = [1.0, 1.0, 1.0, 1.0,  1.0, 1.0], [0.32, 0.32, 0.32, 0.32, 0.32, 0.32]
		mbgr = BatchRegressor(0.1*2, 2)
		epoch!(mbgr, X, Y)
		predict(mbgr, 1.0) ≈ 0.15616 
	end
	
	HTML("<h3 id=q3c> Batch Regressor</h3> 
	<ul><li> batch $(_check_complete(__check_batch_reg))</li> <li>minibatch $(_check_complete(__check_mb)) </li></ul>")
end

# ╔═╡ 7b1d7d43-8613-46dc-9e92-00b91bfd5943
begin
	__check_shr = let
		sgr = StochasticHeuristicRegressor(0.0)
		epoch!(sgr, [1.0, 1.0], [1.0, 1.0])
		predict(sgr, 1.0) * 6 ≈ 5.0
	end
	HTML("<h3 id=q3d> Q3 d: Stochastic Regressor with heuristic $(_check_complete(__check_shr))")
end

# ╔═╡ 375f1a57-50f4-49ba-b089-a38f5315bd81
begin
	__check_bhr_full = let
		X, Y = [1.0, 2.0, 3.0], [1.0, 0.2, 0.1]
		bgr = BatchHeuristicRegressor()
		epoch!(bgr, X, Y)
		batch_test = predict(bgr, 1.0) ≈ 0.36170212765 
	end
		
	__check_bhr_mini = let
		X, Y = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0], [0.32, 0.32, 0.32, 0.32, 0.32, 0.32]
		mbgr = BatchHeuristicRegressor(2)
		epoch!(mbgr, X, Y)
		mb_test = predict(mbgr, 1.0) ≈ 0.319968983713
	end
	HTML("
		<h3 id=q3e> Q3 e: Batch Regressor with heuristic</h3> 
		<ul> 
			<li> Full Batch: $(_check_complete(__check_bhr_full)) </li> 
			<li>Minibatch: $(_check_complete(__check_bhr_mini)) </li>
		</ul>
	")	
end

# ╔═╡ d53a875d-80ed-4d77-b2ea-583f091c7989
let
	q3_a_check = _check_complete(__check_stoch_reg)
	q3_c_check_1 = _check_complete(__check_batch_reg)
	q3_c_check_2 = _check_complete(__check_mb)
	q3_d_check = _check_complete(__check_shr)
	q3_e_check_1 = _check_complete(__check_bhr_full)
	q3_e_check_2 = _check_complete(__check_bhr_mini)
md"""
# Preamble 

In this assignment, we will implement:
- Q3(a,b) [Stochastic Regressor](#q3ab) stochastic regressor $(q3_a_check) 
- Q3(c) [Batch Regressor](#q3c) batch regressor $(q3_c_check_1), and minibatch regressor $(q3_c_check_2)
- Q3(d) [Stochastic Heuristic Regressor](#q3d) $(q3_d_check)
- Q3(e,f) [Batch Heuristic Regressor](#q3e) full batch $(q3_e_check_1), and minibatch $(q3_e_check_2)

You can test your algorithms in [experiment section](#experiment).
"""
end

# ╔═╡ 753cc169-1579-4539-aaa8-7c514e2b3381
md"""
# Plotting Utilities

Below we define two plotting helper functions for using PlotlyJS. You can ignore these if you want. We use them below to compare the algorithms.
"""

# ╔═╡ f80a96ad-9bdf-4ddd-8591-d1a3006e4002
color_scheme = [
    colorant"#44AA99",
    colorant"#332288",
    colorant"#DDCC77",
    colorant"#999933",
    colorant"#CC6677",
    colorant"#AA4499",
    colorant"#117733",
    colorant"#882255",
    colorant"#1E90FF",
]

# ╔═╡ c61c4aaa-0ee7-44a8-8a4c-1119e0c22bde
function boxplot(
		names::Vector{String}, 
		data::Vector{<:AbstractVector};
		col=color_scheme,
		kwargs...)
	colors = 
	traces = GenericTrace{Dict{Symbol, Any}}[]
	for (idx, (name, datum)) in enumerate(zip(names, data))
		tr_bx = box(
			name=name, 
			y=datum, 
			jitter=0.3, 
			marker_color=col[idx])
		push!(traces, tr_bx)
	end
	layout = Layout(; showlegend=false, kwargs...)
	plt = Plot(traces, layout)
end

# ╔═╡ 0825e8cd-26f0-48a3-a146-d6e542f5dd2e
function learning_curve(names, data; err, col = color_scheme, kwargs...)
	trcs = GenericTrace{Dict{Symbol, Any}}[]
	for (idx, (n, d)) in enumerate(zip(names, data))
		terr = scatter(;
			x = vcat(1:length(d), length(d):-1:1),
			y = vcat(d.-err[idx], d.+err[idx]),
			fill="tozerox", 
			fillcolor="rgba($(col[idx].r), $(col[idx].g), $(col[idx].b), 0.2)",
			line_color="transparent")
		t = scatter(;x = 1:length(d), 
				y=d, 
				line_color=col[idx],
				name=n)
		push!(trcs, t)
	end
	layout = Layout(;kwargs...)
	Plot(trcs, layout)
end

# ╔═╡ 5aba1286-42de-4241-ba74-565385043f0b
md"""
# Data

Next we will be looking at the `height_weight.csv` dataset found in the data directory. This dataset provides three features `[sex, height, weight]`. In the following regression task we will be using `height` to predict `weight`, ignoring the `sex` feature.

The next few cells:
- Loads the dataset
- Plots distributions for the `height` and `weight` features seperated by `sex`
- Standardize the set so both `height` and `weight` conform to a standard normal.
- Defines `splitdataframe` which will be used to split the dataframe into training and testing sets.
"""

# ╔═╡ fdca1cde-48ba-4542-b762-b470ee80b1d3
# Read the data from the file in "data/height_weight.csv". DO NOT CHANGE THIS VALUE!
df_height_weight = DataFrame(
	CSV.File(joinpath(@__DIR__, "data/height_weight.csv"), 
	header=["sex", "height", "weight"]));

# ╔═╡ 210f224f-a6aa-4ba7-b560-84ddc18823bf
try 
	identity(df_height_weight)
	md"Successfully loaded dataset ✅"
catch
	Markdown.parse("""Please place datset at `$(joinpath(@__DIR__, "data/height_weight.csv"))`""")
end
	

# ╔═╡ ce1bc6dc-70db-4ad7-b70c-f5c7f6c1129e
df_hw_norm = let
	df = copy(df_height_weight)
	σ_height = sqrt(var(df[!, :height]))
	μ_height = mean(df[!, :height])

	
	df[:, :height] .= (df[!, :height] .- μ_height) ./ σ_height
	
	σ_weight = sqrt(var(df[!, :weight]))
	μ_weight = mean(df[!, :weight])
	df[:, :weight] .= (df[!, :weight] .- μ_weight) ./ σ_weight
	
	df
end

# ╔═╡ 6f4005da-57ff-4161-b4a5-437a1c072dd9
md"""
#### Plot data
Plot a boxplot and violin plot of the height and weight. This can be with the classes `male` and `female` combined or with them separate. 
"""

# ╔═╡ 79dc3541-a75a-4f50-ab0c-2f6522a32eba
plt_hw = let
	df = df_height_weight # For convenience in the bellow code
	nothing
	# plt1 = plot(xlabel="Sex", ylabel="Height", legend=nothing)
	sex_names = ["Male", "Female"]
	get_attr = (sex, attr) -> df[df.sex .== sex, :][!, attr]
	p1 = boxplot(sex_names, 
			get_attr.(sex_names, :height), 
		    yaxis=attr(title="Height"))

	p2 = boxplot(sex_names, 
				 get_attr.(sex_names, :weight), 
				 yaxis=attr(title="Weight"))
	
	p = [p1 p2]
	PlotlyJS.relayout(p, height=400, showlegend=false)

end

# ╔═╡ 4c764fe7-e407-4ed9-9f5a-e2740b9863f6
"""
	splitdataframe(split_to_X_Y::Function, df::DataFrame, test_perc; shuffle = false)
	splitdataframe(df::DataFrame, test_perc; shuffle = false)

Splits a dataframe into test and train sets. Optionally takes a function as the first parameter to split the dataframe into X and Y components for training. This defaults to the `identity` function.
"""
function splitdataframe(split_to_X_Y::Function, df::DataFrame, test_perc; 
		                shuffle = false)
	#= shuffle dataframe. 
	This is innefficient as it makes an entire new dataframe, 
	but fine for the small dataset we have in this notebook. 
	Consider shuffling inplace before calling this function.
	=#
	
	df_shuffle = if shuffle == true
		df[randperm(nrow(df)), :]
	else
		df
	end
	
	# Get train size with percentage of test data.
	train_size = Int(round(size(df,1) * (1 - test_perc)))
	
	dftrain = df_shuffle[1:train_size, :]
	dftest = df_shuffle[(train_size+1):end, :]
	
	split_to_X_Y(dftrain), split_to_X_Y(dftest)
end

# ╔═╡ 790b9f50-7a3b-480c-92f4-33edd967b73d
splitdataframe(df::DataFrame, test_perc; kwargs...) = 
	splitdataframe(identity, df, test_perc; kwargs...)

# ╔═╡ 97dc155f-ef8a-4e55-960c-07fd0f5a6dde
let
	#=
		A do block creates an anonymous function and passes this to the first parameter of the function the do block is decorating.
	=#
	trainset, testset =
		splitdataframe(df_hw_norm, 0.1; shuffle=true) do df
			(X=df[!, :height], Y=df[!, :weight]) # create namedtuple from dataframes
		end
end

# ╔═╡ 168ef51d-dbcd-44c9-8d2e-33e98471962d
md"""
# Training the Models

The following functions are defined as utilities to train and evaluate our models. While hidden below, you can expand these blocks to uncover what is happening. `run_experiment!` is the main function used below in "**Using and Analyzing you Algorithms**".
"""

# ╔═╡ ae23765a-33fd-4a6b-a18c-3081c4382b14
function evaluate(err_func::Function, reg::Regressor, X, Y)
	err_func(predict(reg, X), Y)
end

# ╔═╡ 30177ca7-f9a5-4b97-8967-ffc154a509b0
begin
	evaluate_l1(reg::Regressor, X, Y) = evaluate(l1_error, reg, X, Y)
	evaluate_l2(reg::Regressor, X, Y) = evaluate(l2_error, reg, X, Y)
	evaluate_l∞(reg::Regressor, X, Y) = evaluate(reg, X, Y) do ŷ, y
		maximum(abs.(ŷ .- y))
	end

end

# ╔═╡ 532ed065-1f61-4eb0-9b4f-a6c236ab334d
function train!(ev_func::Function, reg::Regressor, X, Y, num_epochs)
	
	start_error = ev_func(reg, X, Y)
	ret = zeros(typeof(start_error), num_epochs+1)
	ret[1] = start_error
	
	for epoch in 1:num_epochs
		epoch!(reg, X, Y)
		ret[epoch+1] = ev_func(reg, X, Y)
	end
	
	ret
end

# ╔═╡ fbb14b46-51a2-4041-860b-8259f85ef2b7
train!(reg::Regressor, X, Y, num_epochs) = train!(evaluate_l2, reg::Regressor, X, Y, num_epochs)

# ╔═╡ 74b7458e-c654-4b85-9a31-ea575e0aa548
function run_experiment!(reg::Regressor, 
		               	 trainset, 
		                 testset, 
		                 num_epochs)
	
	train_err = train!(reg, trainset.X, trainset.Y, num_epochs)
	test_err = evaluate_l2(reg, testset.X, testset.Y)

	(regressor=reg, train_error=train_err, test_error=test_err)
end

# ╔═╡ 071b7886-ac26-4999-b20d-d63848331ebe
function run_experiment(regressors::Dict{String, Function}, 
						num_runs, 
						num_epochs; 
						seed=10392)

	ret = Dict{String, Any}()
	for (k, reg_init) in regressors
		ret[k] = map(1:num_runs) do r
			Random.seed!(seed+r)
			trainset, testset = splitdataframe(df_hw_norm, 0.1; shuffle=true) do df
				(X=df[!, :height], Y=df[!, :weight]) # create named tuple from DF
			end
			run_experiment!(reg_init(), trainset, testset, num_epochs)
		end
	end

	ret
end

# ╔═╡ 6f2a0d1e-e76b-4326-a350-af49a8fd30e6
html"""<h1 id="experiment"> Using and Analyzing your Algorithms </h1>"""

# ╔═╡ 76d0599d-b893-4cb2-b94d-0e787fd39a74
begin
	
	__s_η = 0.01
	__b_η = 0.01
	__mb_η = 0.01
	__mb_n = 100
	
	Markdown.parse("""

	In this section we will be running and analyzing a small experiment. The goal is to get familiar with analyzing data, plotting learning curves, and comparing different methods. Below we've provided a start with the baselines. Add new initilizors for a Batch update `(η = $(__b_η))`, a Minibatch update `(η = $(__mb_η), n = $(__mb_n))`, and a Stochastic update `(η = $(__s_η))`. Also add their heuristic counterparts. 
		
	As a point of reference: running 
	```julia
	results = run_experiment(regressor_init, 10, 30)
	```
	in the cell below takes roughly `8 seconds` on my machine.

	""")
end

# ╔═╡ 7b80a6fa-4b92-4ade-9c14-785f1a9334d1
md"""
**To run these experiments and draw plots use** $(@bind __run_q3_experiments PlutoUI.CheckBox())
"""

# ╔═╡ 468b1077-6cf4-45d4-a4a6-41b134a6d3d7
regressor_init = Dict(
	# use the keys "Batch", "Stochastic", and "Minibatch", "StochasticHeuristic" "BatchHeuristic", and "MinibatchHeuristic" 
	# This is the actual experiment section
	"Mean"=>()->MeanRegressor(),
	"Random"=>()->RandomRegressor(),
	"Batch"=>()->BatchRegressor(0.01, nothing),
	"Minibatch"=>()->BatchRegressor(0.01, 100),
	"Stochastic"=>()->StochasticRegressor(0.01),
	"BatchHeuristic"=>()->BatchHeuristicRegressor(),
	"MinibatchHeuristic"=>()->BatchHeuristicRegressor(100),
	"StochasticHeuristic"=>()->StochasticHeuristicRegressor()
)

# ╔═╡ ed20ad2c-87ab-4833-ae78-0c1906aa90a6
results = run_experiment(regressor_init, 10, 30)

# ╔═╡ 42ab7e2a-d390-4528-80cd-0e30a1eb2133
begin
	if __run_q3_experiments
		__results_checks = Dict{String, Union{Vector{Bool}, Bool}}(
			"Stochastic"=>false,
			"Minibatch"=>[false, false],
			"Batch"=>false
		)
		
		
		local a = """
		Experiment ran for:
			
		"""
		algs = [
			"Mean",
			"Random", 
			"Stochastic", 
			"Batch", 
			"Minibatch", 
			"StochasticHeuristic", 
			"BatchHeuristic", 
			"MinibatchHeuristic"]
		for k in algs
			if k in keys(results)
				if k == "Batch"
					a *= "- ✅ `$(k)`: "
					__results_checks[k] = results[k][1].regressor.η == __b_η
					a *= """with stepsize= `$(results[k][1].regressor.η)` $(results[k][1].regressor.η == __b_η ? "✅" : "❌" )"""
				elseif k == "Stochastic"
					a *= "- ✅ `$(k)`: "
					__results_checks[k] = results[k][1].regressor.η == __s_η
					a *= """with stepsize= `$(results[k][1].regressor.η)` $(results[k][1].regressor.η == __s_η ? "✅" : "❌" )"""
				elseif k == "Minibatch"
					a *= "- ✅ `$(k)`: "
					__results_checks[k] = [results[k][1].regressor.η == __mb_η, 	
						   				   results[k][1].regressor.b == __mb_n]
					a *= """with stepsize= `$(results[k][1].regressor.η)` $(results[k][1].regressor.η == __mb_η ? "✅" : "❌" ) and batch size = `$(results[k][1].regressor.b)` $(results[k][1].regressor.b == __mb_n ? "✅" : "❌" )"""
				elseif k == "MinibatchHeuristic"
					a *= "- ✅ `$(k)`: "	
					__results_checks[k] = [results[k][1].regressor.b == __mb_n]
					a *= """with batch size = `$(results[k][1].regressor.b)` $(results[k][1].regressor.b == __mb_n ? "✅" : "❌" )"""
				else
					a *= "- ✅ `$(k)`"
				end
			else
				a *= "- ❌ `$(k)`"
			end
			
			a *= "\n\n"
		end
		Markdown.parse(a)
	else
		md"**Run Experiments by clicking above check box.**"
	end
end

# ╔═╡ 3a0a1a93-0ea1-4570-9be6-aa892b515c7b
md"""
The results dictionary is the resulting data from the experiment we run using `regressor_init` as the intializors. You will see the same keys used as in the `regressor_init` dictionary. For each run the experiment returns the final regressor, the training error vector, and the final test error. You can get one of these components for a particular methd using `getindex` and broadcasting:
```julia
getindex.(results["Mean"], :test_error)
```

"""

# ╔═╡ ac703837-ed65-49c6-97ad-5b54e30a680e

let
	if __run_q3_experiments
		# Play with data here! You can explore how to get different values.
		getindex.(results["Mean"], :test_error)
	end
end

# ╔═╡ e3224f78-4cfb-45e7-a1f7-6640051afcd2
md"""
## Learning Curves

Plot the average learning curve with the standard error calculated as

```math
\sigma_{err}(\mathbf{x}) = \sqrt{\frac{\text{Var}(x)}{|x|}}
```

Note that $$\mathbf{x}$$ is a vector over runs, not over epochs.

_Note:_ if you notice one method is dominating the plot, change the axis limits to make sure the methods we are most concerned with (i.e. Stochastic, Batch, and Minibatch) are visible.
"""

# ╔═╡ 6f26e6f8-088c-4578-bf8b-07d38ff53d00
let
	if __run_q3_experiments
		plt = plot()
		ms = collect(keys(results))
		get_train_mean = 
			(d)->mean(getindex.(d, :train_error))
		get_train_std = 
			(d)->sqrt.(var(getindex.(d, 2)) / length(d))
		μs = [get_train_mean(results[m]) for m in ms]
		σs = [get_train_std(results[m]) for m in ms]
		learning_curve(
			ms, 
			μs, 
			err= σs,
			height=400,
			xaxis_range=[0, 31])
	end
end

# ╔═╡ 4f22d201-69bb-4ee1-9393-5058eaffd3d1
md"""
## Final Errors

Finally, we want to compare the final test errors of the different methods. One way to do this is through box plots. See [this great resource](https://towardsdatascience.com/understanding-boxplots-5e2df7bcbd51) to learn how to compare data using a box and whisker plot. In this plot you can ignore the `Random` baseline.
"""

# ╔═╡ 7bc3d05f-9af6-4163-8a31-4143c9606b5b
let
	if __run_q3_experiments
		ms = filter((k)->k ∉ ["Random"], keys(results))
		p1 = boxplot(
			collect(ms), 
			[getindex.(results[m], :test_error) for m in ms], 
			height=400,
			yaxis=attr(title="Test Error"))
		ms = ["Random"]
		p2 = boxplot(
			collect(ms), 
			[getindex.(results[m], :test_error) for m in ms], 
			col=color_scheme[end-1:end],
			height=400)
		
		p = [p1 p2]
		PlotlyJS.relayout(p, height=400, showlegend=false)
	end
end

# ╔═╡ 00000000-0000-0000-0000-000000000001
PLUTO_PROJECT_TOML_CONTENTS = """
[deps]
CSV = "336ed68f-0bac-5ca0-87d4-7b16caf5d00b"
Colors = "5ae59095-9a9b-59fe-a467-6f913c188581"
DataFrames = "a93c6f00-e57d-5684-b7b6-d8193f3e46c0"
Distributions = "31c24e10-a181-5473-b8eb-7969acd0382f"
LinearAlgebra = "37e2e46d-f89d-539d-b4ee-838fcccc9c8e"
PlotlyJS = "f0f68f2c-4968-5e81-91da-67840de0976a"
PlutoUI = "7f904dfe-b85e-4ff6-b463-dae2292396a8"
Random = "9a3f8284-a2c9-5f02-9a11-845980a1fd5c"
Statistics = "10745b16-79ce-11e8-11f9-7d13ad32a3b2"

[compat]
CSV = "~0.8.5"
Colors = "~0.12.8"
DataFrames = "~1.2.2"
Distributions = "~0.25.11"
PlotlyJS = "~0.18.8"
PlutoUI = "~0.7.9"
"""

# ╔═╡ 00000000-0000-0000-0000-000000000002
PLUTO_MANIFEST_TOML_CONTENTS = """
# This file is machine-generated - editing it directly is not advised

[[ArgTools]]
uuid = "0dad84c5-d112-42e6-8d28-ef12dabb789f"

[[Artifacts]]
uuid = "56f22d72-fd6d-98f1-02f0-08ddc0907c33"

[[AssetRegistry]]
deps = ["Distributed", "JSON", "Pidfile", "SHA", "Test"]
git-tree-sha1 = "b25e88db7944f98789130d7b503276bc34bc098e"
uuid = "bf4720bc-e11a-5d0c-854e-bdca1663c893"
version = "0.1.0"

[[Base64]]
uuid = "2a0f44e3-6c83-55bd-87e4-b1978d98bd5f"

[[BinDeps]]
deps = ["Libdl", "Pkg", "SHA", "URIParser", "Unicode"]
git-tree-sha1 = "1289b57e8cf019aede076edab0587eb9644175bd"
uuid = "9e28174c-4ba2-5203-b857-d8d62c4213ee"
version = "1.0.2"

[[Blink]]
deps = ["Base64", "BinDeps", "Distributed", "JSExpr", "JSON", "Lazy", "Logging", "MacroTools", "Mustache", "Mux", "Reexport", "Sockets", "WebIO", "WebSockets"]
git-tree-sha1 = "08d0b679fd7caa49e2bca9214b131289e19808c0"
uuid = "ad839575-38b3-5650-b840-f874b8c74a25"
version = "0.12.5"

[[CSV]]
deps = ["Dates", "Mmap", "Parsers", "PooledArrays", "SentinelArrays", "Tables", "Unicode"]
git-tree-sha1 = "b83aa3f513be680454437a0eee21001607e5d983"
uuid = "336ed68f-0bac-5ca0-87d4-7b16caf5d00b"
version = "0.8.5"

[[ChainRulesCore]]
deps = ["Compat", "LinearAlgebra", "SparseArrays"]
git-tree-sha1 = "bdc0937269321858ab2a4f288486cb258b9a0af7"
uuid = "d360d2e6-b24c-11e9-a2a3-2a2ae2dbcce4"
version = "1.3.0"

[[ColorSchemes]]
deps = ["ColorTypes", "Colors", "FixedPointNumbers", "Random"]
git-tree-sha1 = "6b6f04f93710c71550ec7e16b650c1b9a612d0b6"
uuid = "35d6a980-a343-548e-a6ea-1d62b119f2f4"
version = "3.16.0"

[[ColorTypes]]
deps = ["FixedPointNumbers", "Random"]
git-tree-sha1 = "024fe24d83e4a5bf5fc80501a314ce0d1aa35597"
uuid = "3da002f7-5984-5a60-b8a6-cbb66c0b333f"
version = "0.11.0"

[[Colors]]
deps = ["ColorTypes", "FixedPointNumbers", "Reexport"]
git-tree-sha1 = "417b0ed7b8b838aa6ca0a87aadf1bb9eb111ce40"
uuid = "5ae59095-9a9b-59fe-a467-6f913c188581"
version = "0.12.8"

[[Compat]]
deps = ["Base64", "Dates", "DelimitedFiles", "Distributed", "InteractiveUtils", "LibGit2", "Libdl", "LinearAlgebra", "Markdown", "Mmap", "Pkg", "Printf", "REPL", "Random", "SHA", "Serialization", "SharedArrays", "Sockets", "SparseArrays", "Statistics", "Test", "UUIDs", "Unicode"]
git-tree-sha1 = "344f143fa0ec67e47917848795ab19c6a455f32c"
uuid = "34da2185-b29b-5c13-b0c7-acf172513d20"
version = "3.32.0"

[[CompilerSupportLibraries_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "e66e0078-7015-5450-92f7-15fbd957f2ae"

[[Crayons]]
git-tree-sha1 = "3f71217b538d7aaee0b69ab47d9b7724ca8afa0d"
uuid = "a8cc5b0e-0ffa-5ad4-8c14-923d3ee1735f"
version = "4.0.4"

[[DataAPI]]
git-tree-sha1 = "ee400abb2298bd13bfc3df1c412ed228061a2385"
uuid = "9a962f9c-6df0-11e9-0e5d-c546b8b5ee8a"
version = "1.7.0"

[[DataFrames]]
deps = ["Compat", "DataAPI", "Future", "InvertedIndices", "IteratorInterfaceExtensions", "LinearAlgebra", "Markdown", "Missings", "PooledArrays", "PrettyTables", "Printf", "REPL", "Reexport", "SortingAlgorithms", "Statistics", "TableTraits", "Tables", "Unicode"]
git-tree-sha1 = "d785f42445b63fc86caa08bb9a9351008be9b765"
uuid = "a93c6f00-e57d-5684-b7b6-d8193f3e46c0"
version = "1.2.2"

[[DataStructures]]
deps = ["Compat", "InteractiveUtils", "OrderedCollections"]
git-tree-sha1 = "7d9d316f04214f7efdbb6398d545446e246eff02"
uuid = "864edb3b-99cc-5e75-8d2d-829cb0a9cfe8"
version = "0.18.10"

[[DataValueInterfaces]]
git-tree-sha1 = "bfc1187b79289637fa0ef6d4436ebdfe6905cbd6"
uuid = "e2d170a0-9d28-54be-80f0-106bbe20a464"
version = "1.0.0"

[[Dates]]
deps = ["Printf"]
uuid = "ade2ca70-3891-5945-98fb-dc099432e06a"

[[DelimitedFiles]]
deps = ["Mmap"]
uuid = "8bb1440f-4735-579b-a4ab-409b98df4dab"

[[Distributed]]
deps = ["Random", "Serialization", "Sockets"]
uuid = "8ba89e20-285c-5b6f-9357-94700520ee1b"

[[Distributions]]
deps = ["FillArrays", "LinearAlgebra", "PDMats", "Printf", "QuadGK", "Random", "SparseArrays", "SpecialFunctions", "Statistics", "StatsBase", "StatsFuns"]
git-tree-sha1 = "3889f646423ce91dd1055a76317e9a1d3a23fff1"
uuid = "31c24e10-a181-5473-b8eb-7969acd0382f"
version = "0.25.11"

[[DocStringExtensions]]
deps = ["LibGit2"]
git-tree-sha1 = "a32185f5428d3986f47c2ab78b1f216d5e6cc96f"
uuid = "ffbed154-4ef7-542d-bbb7-c09d3a79fcae"
version = "0.8.5"

[[Downloads]]
deps = ["ArgTools", "LibCURL", "NetworkOptions"]
uuid = "f43a241f-c20a-4ad4-852c-f6b1247861c6"

[[FileWatching]]
uuid = "7b1f6079-737a-58dc-b8bc-7a2ca5c1b5ee"

[[FillArrays]]
deps = ["LinearAlgebra", "Random", "SparseArrays", "Statistics"]
git-tree-sha1 = "8c8eac2af06ce35973c3eadb4ab3243076a408e7"
uuid = "1a297f60-69ca-5386-bcde-b61e274b549b"
version = "0.12.1"

[[FixedPointNumbers]]
deps = ["Statistics"]
git-tree-sha1 = "335bfdceacc84c5cdf16aadc768aa5ddfc5383cc"
uuid = "53c48c17-4a7d-5ca2-90c5-79b7896eea93"
version = "0.8.4"

[[Formatting]]
deps = ["Printf"]
git-tree-sha1 = "8339d61043228fdd3eb658d86c926cb282ae72a8"
uuid = "59287772-0a20-5a39-b81b-1366585eb4c0"
version = "0.4.2"

[[FunctionalCollections]]
deps = ["Test"]
git-tree-sha1 = "04cb9cfaa6ba5311973994fe3496ddec19b6292a"
uuid = "de31a74c-ac4f-5751-b3fd-e18cd04993ca"
version = "0.5.0"

[[Future]]
deps = ["Random"]
uuid = "9fa8497b-333b-5362-9e8d-4d0656e87820"

[[HTTP]]
deps = ["Base64", "Dates", "IniFile", "Logging", "MbedTLS", "NetworkOptions", "Sockets", "URIs"]
git-tree-sha1 = "0fa77022fe4b511826b39c894c90daf5fce3334a"
uuid = "cd3eb016-35fb-5094-929b-558a96fad6f3"
version = "0.9.17"

[[Hiccup]]
deps = ["MacroTools", "Test"]
git-tree-sha1 = "6187bb2d5fcbb2007c39e7ac53308b0d371124bd"
uuid = "9fb69e20-1954-56bb-a84f-559cc56a8ff7"
version = "0.2.2"

[[IniFile]]
deps = ["Test"]
git-tree-sha1 = "098e4d2c533924c921f9f9847274f2ad89e018b8"
uuid = "83e8ac13-25f8-5344-8a64-a9f2b223428f"
version = "0.5.0"

[[InteractiveUtils]]
deps = ["Markdown"]
uuid = "b77e0a4c-d291-57a0-90e8-8db25a27a240"

[[InvertedIndices]]
deps = ["Test"]
git-tree-sha1 = "15732c475062348b0165684ffe28e85ea8396afc"
uuid = "41ab1584-1d38-5bbf-9106-f11c6c58b48f"
version = "1.0.0"

[[IrrationalConstants]]
git-tree-sha1 = "f76424439413893a832026ca355fe273e93bce94"
uuid = "92d709cd-6900-40b7-9082-c6be49f344b6"
version = "0.1.0"

[[IteratorInterfaceExtensions]]
git-tree-sha1 = "a3f24677c21f5bbe9d2a714f95dcd58337fb2856"
uuid = "82899510-4779-5014-852e-03e436cf321d"
version = "1.0.0"

[[JLLWrappers]]
deps = ["Preferences"]
git-tree-sha1 = "642a199af8b68253517b80bd3bfd17eb4e84df6e"
uuid = "692b3bcd-3c85-4b1f-b108-f13ce0eb3210"
version = "1.3.0"

[[JSExpr]]
deps = ["JSON", "MacroTools", "Observables", "WebIO"]
git-tree-sha1 = "bd6c034156b1e7295450a219c4340e32e50b08b1"
uuid = "97c1335a-c9c5-57fe-bc5d-ec35cebe8660"
version = "0.5.3"

[[JSON]]
deps = ["Dates", "Mmap", "Parsers", "Unicode"]
git-tree-sha1 = "8076680b162ada2a031f707ac7b4953e30667a37"
uuid = "682c06a0-de6a-54ab-a142-c8b1cf79cde6"
version = "0.21.2"

[[Kaleido_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "2ef87eeaa28713cb010f9fb0be288b6c1a4ecd53"
uuid = "f7e6163d-2fa5-5f23-b69c-1db539e41963"
version = "0.1.0+0"

[[LaTeXStrings]]
git-tree-sha1 = "f2355693d6778a178ade15952b7ac47a4ff97996"
uuid = "b964fa9f-0449-5b57-a5c2-d3ea65f4040f"
version = "1.3.0"

[[Lazy]]
deps = ["MacroTools"]
git-tree-sha1 = "1370f8202dac30758f3c345f9909b97f53d87d3f"
uuid = "50d2b5c4-7a5e-59d5-8109-a42b560f39c0"
version = "0.15.1"

[[LibCURL]]
deps = ["LibCURL_jll", "MozillaCACerts_jll"]
uuid = "b27032c2-a3e7-50c8-80cd-2d36dbcbfd21"

[[LibCURL_jll]]
deps = ["Artifacts", "LibSSH2_jll", "Libdl", "MbedTLS_jll", "Zlib_jll", "nghttp2_jll"]
uuid = "deac9b47-8bc7-5906-a0fe-35ac56dc84c0"

[[LibGit2]]
deps = ["Base64", "NetworkOptions", "Printf", "SHA"]
uuid = "76f85450-5226-5b5a-8eaa-529ad045b433"

[[LibSSH2_jll]]
deps = ["Artifacts", "Libdl", "MbedTLS_jll"]
uuid = "29816b5a-b9ab-546f-933c-edad1886dfa8"

[[Libdl]]
uuid = "8f399da3-3557-5675-b5ff-fb832c97cbdb"

[[LinearAlgebra]]
deps = ["Libdl", "libblastrampoline_jll"]
uuid = "37e2e46d-f89d-539d-b4ee-838fcccc9c8e"

[[LogExpFunctions]]
deps = ["DocStringExtensions", "IrrationalConstants", "LinearAlgebra"]
git-tree-sha1 = "3d682c07e6dd250ed082f883dc88aee7996bf2cc"
uuid = "2ab3a3ac-af41-5b50-aa03-7779005ae688"
version = "0.3.0"

[[Logging]]
uuid = "56ddb016-857b-54e1-b83d-db4d58db5568"

[[MacroTools]]
deps = ["Markdown", "Random"]
git-tree-sha1 = "3d3e902b31198a27340d0bf00d6ac452866021cf"
uuid = "1914dd2f-81c6-5fcd-8719-6d5c9610ff09"
version = "0.5.9"

[[Markdown]]
deps = ["Base64"]
uuid = "d6f4376e-aef5-505a-96c1-9c027394607a"

[[MbedTLS]]
deps = ["Dates", "MbedTLS_jll", "Random", "Sockets"]
git-tree-sha1 = "1c38e51c3d08ef2278062ebceade0e46cefc96fe"
uuid = "739be429-bea8-5141-9913-cc70e7f3736d"
version = "1.0.3"

[[MbedTLS_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "c8ffd9c3-330d-5841-b78e-0817d7145fa1"

[[Missings]]
deps = ["DataAPI"]
git-tree-sha1 = "4ea90bd5d3985ae1f9a908bd4500ae88921c5ce7"
uuid = "e1d29d7a-bbdc-5cf2-9ac0-f12de2c33e28"
version = "1.0.0"

[[Mmap]]
uuid = "a63ad114-7e13-5084-954f-fe012c677804"

[[MozillaCACerts_jll]]
uuid = "14a3606d-f60d-562e-9121-12d972cd8159"

[[Mustache]]
deps = ["Printf", "Tables"]
git-tree-sha1 = "21d7a05c3b94bcf45af67beccab4f2a1f4a3c30a"
uuid = "ffc61752-8dc7-55ee-8c37-f3e9cdd09e70"
version = "1.0.12"

[[Mux]]
deps = ["AssetRegistry", "Base64", "HTTP", "Hiccup", "Pkg", "Sockets", "WebSockets"]
git-tree-sha1 = "82dfb2cead9895e10ee1b0ca37a01088456c4364"
uuid = "a975b10e-0019-58db-a62f-e48ff68538c9"
version = "0.7.6"

[[NetworkOptions]]
uuid = "ca575930-c2e3-43a9-ace4-1e988b2c1908"

[[Observables]]
git-tree-sha1 = "fe29afdef3d0c4a8286128d4e45cc50621b1e43d"
uuid = "510215fc-4207-5dde-b226-833fc4488ee2"
version = "0.4.0"

[[OpenBLAS_jll]]
deps = ["Artifacts", "CompilerSupportLibraries_jll", "Libdl"]
uuid = "4536629a-c528-5b80-bd46-f80d51c5b363"

[[OpenSpecFun_jll]]
deps = ["Artifacts", "CompilerSupportLibraries_jll", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "13652491f6856acfd2db29360e1bbcd4565d04f1"
uuid = "efe28fd5-8261-553b-a9e1-b2916fc3738e"
version = "0.5.5+0"

[[OrderedCollections]]
git-tree-sha1 = "85f8e6578bf1f9ee0d11e7bb1b1456435479d47c"
uuid = "bac558e1-5e72-5ebc-8fee-abe8a469f55d"
version = "1.4.1"

[[PDMats]]
deps = ["LinearAlgebra", "SparseArrays", "SuiteSparse"]
git-tree-sha1 = "4dd403333bcf0909341cfe57ec115152f937d7d8"
uuid = "90014a1f-27ba-587c-ab20-58faa44d9150"
version = "0.11.1"

[[Parameters]]
deps = ["OrderedCollections", "UnPack"]
git-tree-sha1 = "34c0e9ad262e5f7fc75b10a9952ca7692cfc5fbe"
uuid = "d96e819e-fc66-5662-9728-84c9c7592b0a"
version = "0.12.3"

[[Parsers]]
deps = ["Dates"]
git-tree-sha1 = "bfd7d8c7fd87f04543810d9cbd3995972236ba1b"
uuid = "69de0a69-1ddd-5017-9359-2bf0b02dc9f0"
version = "1.1.2"

[[Pidfile]]
deps = ["FileWatching", "Test"]
git-tree-sha1 = "1be8660b2064893cd2dae4bd004b589278e4440d"
uuid = "fa939f87-e72e-5be4-a000-7fc836dbe307"
version = "1.2.0"

[[Pkg]]
deps = ["Artifacts", "Dates", "Downloads", "LibGit2", "Libdl", "Logging", "Markdown", "Printf", "REPL", "Random", "SHA", "Serialization", "TOML", "Tar", "UUIDs", "p7zip_jll"]
uuid = "44cfe95a-1eb2-52ea-b672-e2afdf69b78f"

[[PlotlyBase]]
deps = ["ColorSchemes", "Dates", "DelimitedFiles", "DocStringExtensions", "JSON", "LaTeXStrings", "Logging", "Parameters", "Pkg", "REPL", "Requires", "Statistics", "UUIDs"]
git-tree-sha1 = "180d744848ba316a3d0fdf4dbd34b77c7242963a"
uuid = "a03496cd-edff-5a9b-9e67-9cda94a718b5"
version = "0.8.18"

[[PlotlyJS]]
deps = ["Base64", "Blink", "DelimitedFiles", "JSExpr", "JSON", "Kaleido_jll", "Markdown", "Pkg", "PlotlyBase", "REPL", "Reexport", "Requires", "WebIO"]
git-tree-sha1 = "53d6325e14d3bdb85fd387a085075f36082f35a3"
uuid = "f0f68f2c-4968-5e81-91da-67840de0976a"
version = "0.18.8"

[[PlutoUI]]
deps = ["Base64", "Dates", "InteractiveUtils", "JSON", "Logging", "Markdown", "Random", "Reexport", "Suppressor"]
git-tree-sha1 = "44e225d5837e2a2345e69a1d1e01ac2443ff9fcb"
uuid = "7f904dfe-b85e-4ff6-b463-dae2292396a8"
version = "0.7.9"

[[PooledArrays]]
deps = ["DataAPI", "Future"]
git-tree-sha1 = "cde4ce9d6f33219465b55162811d8de8139c0414"
uuid = "2dfb63ee-cc39-5dd5-95bd-886bf059d720"
version = "1.2.1"

[[Preferences]]
deps = ["TOML"]
git-tree-sha1 = "00cfd92944ca9c760982747e9a1d0d5d86ab1e5a"
uuid = "21216c6a-2e73-6563-6e65-726566657250"
version = "1.2.2"

[[PrettyTables]]
deps = ["Crayons", "Formatting", "Markdown", "Reexport", "Tables"]
git-tree-sha1 = "0d1245a357cc61c8cd61934c07447aa569ff22e6"
uuid = "08abe8d2-0d0c-5749-adfa-8a2ac140af0d"
version = "1.1.0"

[[Printf]]
deps = ["Unicode"]
uuid = "de0858da-6303-5e67-8744-51eddeeeb8d7"

[[QuadGK]]
deps = ["DataStructures", "LinearAlgebra"]
git-tree-sha1 = "12fbe86da16df6679be7521dfb39fbc861e1dc7b"
uuid = "1fd47b50-473d-5c70-9696-f719f8f3bcdc"
version = "2.4.1"

[[REPL]]
deps = ["InteractiveUtils", "Markdown", "Sockets", "Unicode"]
uuid = "3fa0cd96-eef1-5676-8a61-b3b8758bbffb"

[[Random]]
deps = ["SHA", "Serialization"]
uuid = "9a3f8284-a2c9-5f02-9a11-845980a1fd5c"

[[Reexport]]
git-tree-sha1 = "5f6c21241f0f655da3952fd60aa18477cf96c220"
uuid = "189a3867-3050-52da-a836-e630ba90ab69"
version = "1.1.0"

[[Requires]]
deps = ["UUIDs"]
git-tree-sha1 = "838a3a4188e2ded87a4f9f184b4b0d78a1e91cb7"
uuid = "ae029012-a4dd-5104-9daa-d747884805df"
version = "1.3.0"

[[Rmath]]
deps = ["Random", "Rmath_jll"]
git-tree-sha1 = "bf3188feca147ce108c76ad82c2792c57abe7b1f"
uuid = "79098fc4-a85e-5d69-aa6a-4863f24498fa"
version = "0.7.0"

[[Rmath_jll]]
deps = ["Artifacts", "JLLWrappers", "Libdl", "Pkg"]
git-tree-sha1 = "68db32dff12bb6127bac73c209881191bf0efbb7"
uuid = "f50d1b31-88e8-58de-be2c-1cc44531875f"
version = "0.3.0+0"

[[SHA]]
uuid = "ea8e919c-243c-51af-8825-aaa63cd721ce"

[[SentinelArrays]]
deps = ["Dates", "Random"]
git-tree-sha1 = "a3a337914a035b2d59c9cbe7f1a38aaba1265b02"
uuid = "91c51154-3ec4-41a3-a24f-3f23e20d615c"
version = "1.3.6"

[[Serialization]]
uuid = "9e88b42a-f829-5b0c-bbe9-9e923198166b"

[[SharedArrays]]
deps = ["Distributed", "Mmap", "Random", "Serialization"]
uuid = "1a1011a3-84de-559e-8e89-a11a2f7dc383"

[[Sockets]]
uuid = "6462fe0b-24de-5631-8697-dd941f90decc"

[[SortingAlgorithms]]
deps = ["DataStructures"]
git-tree-sha1 = "b3363d7460f7d098ca0912c69b082f75625d7508"
uuid = "a2af1166-a08f-5f64-846c-94a0d3cef48c"
version = "1.0.1"

[[SparseArrays]]
deps = ["LinearAlgebra", "Random"]
uuid = "2f01184e-e22b-5df5-ae63-d93ebab69eaf"

[[SpecialFunctions]]
deps = ["ChainRulesCore", "LogExpFunctions", "OpenSpecFun_jll"]
git-tree-sha1 = "a322a9493e49c5f3a10b50df3aedaf1cdb3244b7"
uuid = "276daf66-3868-5448-9aa4-cd146d93841b"
version = "1.6.1"

[[Statistics]]
deps = ["LinearAlgebra", "SparseArrays"]
uuid = "10745b16-79ce-11e8-11f9-7d13ad32a3b2"

[[StatsAPI]]
git-tree-sha1 = "1958272568dc176a1d881acb797beb909c785510"
uuid = "82ae8749-77ed-4fe6-ae5f-f523153014b0"
version = "1.0.0"

[[StatsBase]]
deps = ["DataAPI", "DataStructures", "LinearAlgebra", "Missings", "Printf", "Random", "SortingAlgorithms", "SparseArrays", "Statistics", "StatsAPI"]
git-tree-sha1 = "fed1ec1e65749c4d96fc20dd13bea72b55457e62"
uuid = "2913bbd2-ae8a-5f71-8c99-4fb6c76f3a91"
version = "0.33.9"

[[StatsFuns]]
deps = ["IrrationalConstants", "LogExpFunctions", "Reexport", "Rmath", "SpecialFunctions"]
git-tree-sha1 = "20d1bb720b9b27636280f751746ba4abb465f19d"
uuid = "4c63d2b9-4356-54db-8cca-17b64c39e42c"
version = "0.9.9"

[[SuiteSparse]]
deps = ["Libdl", "LinearAlgebra", "Serialization", "SparseArrays"]
uuid = "4607b0f0-06f3-5cda-b6b1-a6196a1729e9"

[[Suppressor]]
git-tree-sha1 = "a819d77f31f83e5792a76081eee1ea6342ab8787"
uuid = "fd094767-a336-5f1f-9728-57cf17d0bbfb"
version = "0.2.0"

[[TOML]]
deps = ["Dates"]
uuid = "fa267f1f-6049-4f14-aa54-33bafae1ed76"

[[TableTraits]]
deps = ["IteratorInterfaceExtensions"]
git-tree-sha1 = "c06b2f539df1c6efa794486abfb6ed2022561a39"
uuid = "3783bdb8-4a98-5b6b-af9a-565f29a5fe9c"
version = "1.0.1"

[[Tables]]
deps = ["DataAPI", "DataValueInterfaces", "IteratorInterfaceExtensions", "LinearAlgebra", "TableTraits", "Test"]
git-tree-sha1 = "d0c690d37c73aeb5ca063056283fde5585a41710"
uuid = "bd369af6-aec1-5ad0-b16a-f7cc5008161c"
version = "1.5.0"

[[Tar]]
deps = ["ArgTools", "SHA"]
uuid = "a4e569a6-e804-4fa4-b0f3-eef7a1d5b13e"

[[Test]]
deps = ["InteractiveUtils", "Logging", "Random", "Serialization"]
uuid = "8dfed614-e22c-5e08-85e1-65c5234f0b40"

[[URIParser]]
deps = ["Unicode"]
git-tree-sha1 = "53a9f49546b8d2dd2e688d216421d050c9a31d0d"
uuid = "30578b45-9adc-5946-b283-645ec420af67"
version = "0.4.1"

[[URIs]]
git-tree-sha1 = "97bbe755a53fe859669cd907f2d96aee8d2c1355"
uuid = "5c2747f8-b7ea-4ff2-ba2e-563bfd36b1d4"
version = "1.3.0"

[[UUIDs]]
deps = ["Random", "SHA"]
uuid = "cf7118a7-6976-5b1a-9a39-7adc72f591a4"

[[UnPack]]
git-tree-sha1 = "387c1f73762231e86e0c9c5443ce3b4a0a9a0c2b"
uuid = "3a884ed6-31ef-47d7-9d2a-63182c4928ed"
version = "1.0.2"

[[Unicode]]
uuid = "4ec0a83e-493e-50e2-b9ac-8f72acf5a8f5"

[[WebIO]]
deps = ["AssetRegistry", "Base64", "Distributed", "FunctionalCollections", "JSON", "Logging", "Observables", "Pkg", "Random", "Requires", "Sockets", "UUIDs", "WebSockets", "Widgets"]
git-tree-sha1 = "5fe32e4086d49f7ab9b087296742859f3ae6d62a"
uuid = "0f1e0344-ec1d-5b48-a673-e5cf874b6c29"
version = "0.8.16"

[[WebSockets]]
deps = ["Base64", "Dates", "HTTP", "Logging", "Sockets"]
git-tree-sha1 = "f91a602e25fe6b89afc93cf02a4ae18ee9384ce3"
uuid = "104b5d7c-a370-577a-8038-80a2059c5097"
version = "1.5.9"

[[Widgets]]
deps = ["Colors", "Dates", "Observables", "OrderedCollections"]
git-tree-sha1 = "80661f59d28714632132c73779f8becc19a113f2"
uuid = "cc8bc4a8-27d6-5769-a93b-9d913e69aa62"
version = "0.6.4"

[[Zlib_jll]]
deps = ["Libdl"]
uuid = "83775a58-1f1d-513f-b197-d71354ab007a"

[[libblastrampoline_jll]]
deps = ["Artifacts", "Libdl", "OpenBLAS_jll"]
uuid = "8e850b90-86db-534c-a0d3-1478176c7d93"

[[nghttp2_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "8e850ede-7688-5339-a07c-302acd2aaf8d"

[[p7zip_jll]]
deps = ["Artifacts", "Libdl"]
uuid = "3f19e933-33d8-53b3-aaab-bd5110c3b7a0"
"""

# ╔═╡ Cell order:
# ╟─6fef79fe-aa3c-497b-92d3-6ddd87b6c26d
# ╠═a7b272bc-e8f5-11eb-38c6-8f61fea9941c
# ╠═ac9b31de-a4f0-4cb6-82a7-890436ff1bed
# ╠═131466aa-d30d-4f75-a99f-13f47e1c7956
# ╠═7f8a82ea-1690-490c-a8bc-3c1f9556af2e
# ╟─5924a9a5-88c7-4751-af77-4a217dfdc15f
# ╟─72d6b53f-f20d-4c05-abd4-d6fb4c997795
# ╠═9d38b22c-5893-4587-9c33-82d846fd0728
# ╟─e10ffb7d-ede9-4cec-8908-836cd1ee3ae1
# ╠═a095e895-cabb-4e3a-a027-ddd6aa356e41
# ╟─d53a875d-80ed-4d77-b2ea-583f091c7989
# ╟─c8ceaf7a-d740-4123-8b26-e652dd17a86e
# ╠═b3edc600-6670-4ae6-af7a-337e0600390e
# ╟─ac180a6e-009f-4db2-bdfc-a9011dc5eb7b
# ╟─999ce1f2-fd11-4584-9c2e-bb79585d11f7
# ╠═80f4168d-c5a4-41ef-a542-4e6d7772aa80
# ╠═65505c23-5096-4621-b3d4-d8d5ed339ad5
# ╠═67a8db8a-365f-4912-a058-d61648ac096e
# ╟─7e083787-7a16-4064-97d9-8e2fca6ed686
# ╟─5d5fe2d8-98ea-4ee1-b2e2-354eefaf8f77
# ╠═eb152f9f-908f-4e5b-9642-f7314eb66e09
# ╟─a7b7bdaf-f4fd-40fb-a704-a584a93920f2
# ╠═0a90ceec-fc50-41fb-bf8b-e96ed677a5c3
# ╟─76c5de40-a0de-4d90-82f0-4f6c941499d9
# ╟─e8f3981d-545c-4e35-9066-69fa4f78dbce
# ╠═6ef4d625-fdf6-4f11-81f1-b242b2195e8b
# ╠═579da3a3-0e4e-4ba2-8f44-2657229994e3
# ╠═67274c5b-2ee3-4865-a5e1-799db543d0c7
# ╠═57c4dbf9-b7ba-44e9-8120-0e1a85896976
# ╟─06be63f6-577c-4117-8dac-bba8bba8b9ce
# ╟─add25a22-7423-4077-b0b0-f71eed2d2e20
# ╠═b0160dc0-35af-4750-9ac6-d9018ce89ea9
# ╠═adedf31d-945f-491a-912a-bef9b03f6665
# ╟─debd1501-2fae-431c-85cf-bea549a4bba4
# ╟─9b69b182-ee8d-485e-ad6b-1f868874652c
# ╠═4d422513-34e3-4d04-8ff1-c5165d953342
# ╠═e5d78429-55f4-4a1a-831f-dbcb53c6a0f6
# ╟─2aa8bd30-4b65-4e89-9dd5-5333efbbda3f
# ╟─7b1d7d43-8613-46dc-9e92-00b91bfd5943
# ╠═51bc41b4-b27f-4e60-8dba-70783c60c1c2
# ╠═00fcafa5-308c-4895-ad52-961772348125
# ╟─375f1a57-50f4-49ba-b089-a38f5315bd81
# ╠═7d6027af-6bfe-4cca-8b81-3e8fb12a6f79
# ╠═297b3bd1-29cc-4a96-bf2c-6305f8d375e4
# ╟─753cc169-1579-4539-aaa8-7c514e2b3381
# ╟─f80a96ad-9bdf-4ddd-8591-d1a3006e4002
# ╠═c61c4aaa-0ee7-44a8-8a4c-1119e0c22bde
# ╠═0825e8cd-26f0-48a3-a146-d6e542f5dd2e
# ╟─5aba1286-42de-4241-ba74-565385043f0b
# ╠═fdca1cde-48ba-4542-b762-b470ee80b1d3
# ╟─210f224f-a6aa-4ba7-b560-84ddc18823bf
# ╟─ce1bc6dc-70db-4ad7-b70c-f5c7f6c1129e
# ╟─6f4005da-57ff-4161-b4a5-437a1c072dd9
# ╠═79dc3541-a75a-4f50-ab0c-2f6522a32eba
# ╟─4c764fe7-e407-4ed9-9f5a-e2740b9863f6
# ╟─790b9f50-7a3b-480c-92f4-33edd967b73d
# ╠═97dc155f-ef8a-4e55-960c-07fd0f5a6dde
# ╟─168ef51d-dbcd-44c9-8d2e-33e98471962d
# ╟─ae23765a-33fd-4a6b-a18c-3081c4382b14
# ╟─30177ca7-f9a5-4b97-8967-ffc154a509b0
# ╟─532ed065-1f61-4eb0-9b4f-a6c236ab334d
# ╟─fbb14b46-51a2-4041-860b-8259f85ef2b7
# ╟─74b7458e-c654-4b85-9a31-ea575e0aa548
# ╟─071b7886-ac26-4999-b20d-d63848331ebe
# ╟─6f2a0d1e-e76b-4326-a350-af49a8fd30e6
# ╟─76d0599d-b893-4cb2-b94d-0e787fd39a74
# ╟─7b80a6fa-4b92-4ade-9c14-785f1a9334d1
# ╟─42ab7e2a-d390-4528-80cd-0e30a1eb2133
# ╠═468b1077-6cf4-45d4-a4a6-41b134a6d3d7
# ╠═ed20ad2c-87ab-4833-ae78-0c1906aa90a6
# ╟─3a0a1a93-0ea1-4570-9be6-aa892b515c7b
# ╠═ac703837-ed65-49c6-97ad-5b54e30a680e
# ╟─e3224f78-4cfb-45e7-a1f7-6640051afcd2
# ╟─6f26e6f8-088c-4578-bf8b-07d38ff53d00
# ╟─4f22d201-69bb-4ee1-9393-5058eaffd3d1
# ╟─7bc3d05f-9af6-4163-8a31-4143c9606b5b
# ╟─00000000-0000-0000-0000-000000000001
# ╟─00000000-0000-0000-0000-000000000002
